#!/usr/bin/python3

#	cve-manager : CVE management tool
#	Copyright (C) 2017-2022 Alexey Appolonov
#
#	This program is free software: you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation, either version 3 of the License, or
#	(at your option) any later version.
#
#	This program is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with this program.  If not, see <http://www.gnu.org/licenses/>.

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

import re
import argparse
from ax.purefunctions    import ReplaceNotAlfa
from cve_manager.desc    import MAPPER
from cve_manager.defines import DATA_SOURCES
from cve_manager.common  import NewArgParser as NewCommonArgParser

DELIMITER = '_'

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Parsers of arguments of every module that performs a matching

def BlankArgParser():

	return argparse.ArgumentParser(description=MAPPER)


def NewArgParser(base=None, ptype=''):

	if base:
		argparser = base
	else:
		argparser = BlankArgParser()

	argparser.add_argument(
		'-d', '--data_sources',
		metavar='ISSUES_TYPE', type=str, nargs='+',
		choices=DATA_SOURCES, default=DATA_SOURCES,
		help='Sources of vulnerability lists whose products will be mapped '
		f'to package names ({" or ".join(DATA_SOURCES)}, all by default)'
		)

	filtered_ptype = [c for c in ptype if c in ('b', 'p')]

	argparser = NewCommonArgParser(base=argparser, ptype=filtered_ptype)
	argparser.add_argument(
		'-n', '--names',
		metavar='CUSTOM_NAME', type=str, nargs='+', default = [],
		help='Custom names that names of products will be mapped to '
		'(for debugging purposes, resulting map will not be sent to VUL DB)'
		)
	argparser.add_argument(
		'--noupdate',
		action='store_true',
		help='Do not send results to DB'
		)

	return NewCommonArgParser(base=argparser, ptype='m')

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Make all symbols lowercase, replace all delimiters with a symbol defined in
# the constant "DELIMITER" and remove all delimiters at the end

def FmtName(name):

	return re.sub(
		r'_+$',
		'',
		ReplaceNotAlfa(
			name.lower(), DELIMITER, allow_in_versions=True, exceptions={'+'}
			)
		)

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Check compatibility of specified groups

def CompatibleGroups(a, b):

	return a == b or a == 0 or b == 0

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Sort matches by a crit value and forming a string of a particular format

def MatchesStr(scores_for_products, vendor=None):

	sorted_matches = \
		sorted(scores_for_products.items(), key=lambda kv: kv[1], reverse=True)
	prefix = f'{vendor}:' if vendor else ''

	return '  '.join([f"{prefix}{p} {n}" for p, n in sorted_matches if p])

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Form a CPE-style string out of (vendor, product) tuple

def FormCPEStyleStr(vendor, product):

	return f'{vendor}:{product}' if vendor else product
