#!/usr/bin/python3

#	cve-manager : CVE management tool
#	Copyright (C) 2017-2022 Alexey Appolonov
#
#	This program is free software: you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation, either version 3 of the License, or
#	(at your option) any later version.
#
#	This program is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with this program.  If not, see <http://www.gnu.org/licenses/>.

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

from re                            import match
from collections                   import defaultdict
from cve_manager.common            import MATCH_VENDOR_PRODUCT
from cve_manager.knowledge_handler import KnowledgeHandler

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Class that forms and stores a dict of related CPEs

class PrescribedCPERelatives(KnowledgeHandler):

	def __init__(self, conf_dir):

		super().__init__(conf_dir, 'cpe-relatives.txt')

	def __WarnMsg(self, path, line):

		return f'Wrong format of the "{self.file_path}" file: {line}'

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Read the list of prescribed CPE relatives and form a dict of related CPEs
	# {(<vendor_a>, <product_a>: {(<vendor_b>, <product_b>), ...}}

	def Read(self):

		ok, msg = super().Check()
		if not ok:
			return None, msg

		prescriptions = defaultdict(set)
		warn = []

		with open(self.file_path, 'r') as f:

			for line in f.readlines():

				line = line.strip()

				if not line:
					continue

				cols = set(line.split())
				if len(cols) < 2:
					warn.append(self.__WarnMsg(self.file_path, line))
					continue

				relatives = set()

				for col in cols:
					m = match(fr'^{MATCH_VENDOR_PRODUCT}$', col)
					if not m:
						relatives.clear()
						warn.append(self.__WarnMsg(self.file_path, line))
						break
					vendor = m.group('vendor')[:-1] if m.group('vendor') else ''
					product = m.group('product')
					relatives.add((vendor, product))

				for relative in relatives:
					prescriptions[relative] |= relatives - {relative}

		return prescriptions, '; '.join(warn)
