#!/usr/bin/python3

#	cve-manager : CVE management tool
#	Copyright (C) 2017-2022 Alexey Appolonov
#
#	This program is free software: you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation, either version 3 of the License, or
#	(at your option) any later version.
#
#	This program is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with this program.  If not, see <http://www.gnu.org/licenses/>.

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

import re
from collections                   import defaultdict
from cve_manager.db                import DB, Placeholders
from cve_manager.knowledge_handler import KnowledgeHandler
from cve_manager.url               import ParseAndJoinURL
from cpe_map.common                import DELIMITER, FmtName

G_TABLE = 'groups_of_packages_and_products'
G_ID = 'id'
G_PACKAGE_PREFIX = 'package_prefix'
G_PRODUCT_PREFIX = 'product_prefix'
G_PRODUCT_SUFFIX = 'product_suffix'
G_COLS = [G_ID, G_PACKAGE_PREFIX, G_PRODUCT_PREFIX, G_PRODUCT_SUFFIX]

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Combine multiple group numbers into one (-1 if not possible)

def CombineGroups(groups):

	comb = 0

	for group in groups:
		if group < 0:
			return -1
		if comb == 0:
			comb = group
		elif group != comb:
			return -1

	return comb

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Check that one of the two URLs contains the other

def OneUrlIncludesAnother(a, b):

	if len(a) < len(b):
		a, b = b, a

	i = a.find(b)

	return i >= 0 and (i == 0 or a[i - 1] == '.') and \
		((len(a) - i - len(b)) == 0 or a[i + len(b)] == '/')

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

class Groups(KnowledgeHandler):

	PARAMS_Q = 5                # Presence in the "groups" table:
	PARAMS_PACKAGE_PREFIXES = 0 # [v]
	PARAMS_VENDORS = 1          # [-]
	PARAMS_PRODUCT_PREFIXES = 2 # [v]
	PARAMS_PRODUCT_SUFFIXES = 3 # [v]
	PARAMS_URLS = 4             # [-]

	def __init__(self, conf_dir, mysql_config, printer):

		super().__init__(conf_dir, 'groups.csv')

		self.printer = printer
		self.mysql_config = mysql_config
		self.groups = {}
		self.ready = False

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Read the special file if it hasn't been read yet and access formed dict
	# "self.groups"

	def __AccessGroups(self, param_index):

		if not self.ready and not self.__Prepare():
			return None

		return {field: group
			for group, params in self.groups.items()
				for fields in params[param_index]
					for field in fields.split()}

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Read the special file and form a dict of groups that describe high level
	# relationship between packages and products

	def __Read(self):

		self.printer.LineBegin(f'Reading the "{self.file_path}" list')

		ok, msg = super().Check()
		if not ok:
			self.printer.Err(msg)
			return False

		self.groups.clear()
		header = True
		warn = [] if not msg else [msg]

		with open(self.file_path, 'r') as f:
			for line_n, line in enumerate(f.readlines()):
				comment_pos = line.find('#')
				if comment_pos >= 0:
					line = line[:comment_pos]
				cols = [col.strip() for col in line.split(', ')]
				if len(cols) < (self.PARAMS_Q + 1):
					warn.append(f'Wrong format in "{self.file_path}", '
						f'line {line_n} -- skipping')
					continue
				if header:
					header = False
					continue
				params = []
				for i, col in enumerate(cols[1:]):
					param = set()
					if col != '-':
						for el in col.split():
							if i == self.PARAMS_URLS:
								col = ParseAndJoinURL(el)
							param.add(col)
					params.append(list(param))
					params[-1].sort(key = lambda x: len(x), reverse=True)
				self.groups[cols[0]] = params

		# Using indexes of groups instead of group names
		self.groups = {(i + 1): params
			for i, params in enumerate(self.groups.values())}

		self.ready = True

		if warn:
			self.printer.Warn('; '.join(warn))
		else:
			self.printer.Success()

		return True

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Read the special file, recreate the "groups" table and send the data
	# to the "groups" table

	def Prepare(self):

		if not self.__Read():
			return False

		self.printer.LineBegin(f'Preparing the "{G_TABLE}" table')

		db = DB(self.printer)
		if not db.Connect(self.mysql_config):
			return False

		columns = [f'{G_ID} TINYINT'] + \
			[f'{col} TEXT' for col in G_COLS[1:4]] + \
			[f'CONSTRAINT pk_{G_TABLE} PRIMARY KEY ({G_ID})']

		try:
			if not db.RecreateTable(G_TABLE, columns):
				return False
		except MySQLError as err:
			self.printer.Err(f'{err}')
			return False

		query = f"INSERT INTO {G_TABLE} ({', '.join(G_COLS)}) " \
			f"VALUES ({Placeholders(len(G_COLS))})"

		data = []
		for group, params in self.groups.items():
			entry = [group]
			for i, param in enumerate(params):
				if i in (self.PARAMS_PACKAGE_PREFIXES,
						self.PARAMS_PRODUCT_PREFIXES,
						self.PARAMS_PRODUCT_SUFFIXES):
					entry.append(' '.join(param))
			data.append(entry)

		if not data:
			self.printer.Err('No groups')
			return False

		if db.Cursor(query, data=data) == None:
			return False

		self.printer.Success()

		return True

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Determine group of a specified package

	def CheckPackage(self, package_name):

		prefixes = self.__AccessGroups(self.PARAMS_PACKAGE_PREFIXES)

		for prefix, group in prefixes.items():
			if package_name.startswith(prefix + '-'):
				return group

		return 0

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Determine group of a specified CPE

	def CheckProduct(self, vendor, product, refs):

		groups = []
		product = FmtName(product)

		# For every type of group param
		for i in range(1, self.PARAMS_Q):

			d = self.__AccessGroups(i)

			if i == self.PARAMS_VENDORS:
				for vendor_from_file, group_from_file in d.items():
					m = re.match(r'^r\"(.*)\"$', vendor_from_file)
					if m:
						if re.search(fr'{m.group(1)}', vendor):
							groups.append(group_from_file)
					else:
						if vendor == vendor_from_file:
							groups.append(group_from_file)

			elif i == self.PARAMS_PRODUCT_PREFIXES:
				for prefix, group_from_file in d.items():
					if product.startswith(prefix + DELIMITER):
						groups.append(group_from_file)

			elif i == self.PARAMS_PRODUCT_SUFFIXES:
				for suffix, group_from_file in d.items():
					if product.endswith(DELIMITER + suffix):
						groups.append(group_from_file)

			elif i == self.PARAMS_URLS:	
				for ref in refs:
					url = ParseAndJoinURL(ref)
					for url_from_file, group_from_file in d.items():
						if OneUrlIncludesAnother(url, url_from_file):
							groups.append(group_from_file)

		return CombineGroups(groups)
