#!/usr/bin/python3

#	cve-manager : CVE management tool
#	Copyright (C) 2017-2022 Alexey Appolonov
#
#	This program is free software: you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation, either version 3 of the License, or
#	(at your option) any later version.
#
#	This program is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with this program.  If not, see <http://www.gnu.org/licenses/>.

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

from mysql.connector      import Error as MySQLError
from collections          import defaultdict
from cve_manager.mediator import ProtoMediator
from cve_manager.common   import SplitData
from cve_manager.db       import DB, Placeholders
from cve_manager.defines  import DATA_SOURCES, M_URL, M_FIXES
from cpe_map.groups       import G_TABLE, G_COLS, G_ID, CombineGroups
from cpe_map.name_conv    import NameConverter, NAME_NEEDED_PARAMS

EMPTY_DICT_MSG = 'Empty dict'

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# An interface between modules that perform matching and a DB

class Mediator(ProtoMediator):

	def __init__(self, type_of_matching, mysql_config, conf_dir, printer,
			target_is_bin, no_update):

		super().__init__(mysql_config, printer)

		self.type_of_matching = type_of_matching
		self.target_is_bin = target_is_bin
		self.no_update = no_update
		self.conf_dir = conf_dir
		self.groups = None
		self.name_conv = None

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Get a dict of group parameters of packages/products

	def __QueryGroups(self):

		self.printer.LineBegin('Get a dict of group parameters of packages/'
			'products')

		db = super().NewDB()
		if not db:
			return None

		cursor = db.Select(G_COLS, G_TABLE)
		if not cursor:
			if cursor != None:
				self.printer.Warn(EMPTY_DICT_MSG)
				return {}
			return None

		groups = {}
		g_id_index = 0

		for i in range(len(G_COLS)):
			if G_COLS[i] == G_ID:
				g_id_index = i
				break

		for entry in cursor:
			groups[entry[g_id_index]] = {G_COLS[i]: entry[i].split()
				for i in range(len(G_COLS)) if i != g_id_index}

		if not groups:
			self.printer.Warn(EMPTY_DICT_MSG)
		else:
			self.printer.Success()

		return groups

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Initialize auxiliary objects if they are not initialized

	def __InitAuxObjects(self):

		if self.groups and self.name_conv:
			return True

		self.groups = self.__QueryGroups()
		if self.groups == None:
			return False

		self.name_conv = NameConverter(self.conf_dir, self.groups)
		return True

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Get a dict of src or bin package names

	def QueryPackages(self, selection=[], blind_selection=[], n_parts=-1):

		if not self.__InitAuxObjects():
			return None

		self.printer.LineBegin('Getting a dict of united '
			f'{"bin" if self.target_is_bin else "src"} package names')

		name_needed_params = NAME_NEEDED_PARAMS.get(self.type_of_matching)
		if not name_needed_params:
			self.printer.Err(
				f'Unknown type of matching "{self.type_of_matching}"')
			return None

		if blind_selection:
			self.printer.Success('Using given names')
			packages = \
				{name: self.name_conv.Do(name, 0, True, name_needed_params)
					for name in blind_selection}
			return [packages] if n_parts > 0 else packages

		# Connecting to a CVE database
		db = super().NewDB()
		if not db:
			return None

		# Getting package names
		table = f'{"bin" if self.target_is_bin else "src"}_united'
		cols  = ['name', 'group_id']
		if selection:
			cond = ("src" if self.target_is_bin else "name" ) + ' in ' \
				f'({Placeholders(len(selection))})'
			data = [tuple(selection)]
		else:
			cond = None
			data = []
		packages_cursor = db.Select(cols, table, cond, data)
		if not packages_cursor:
			if packages_cursor != None:
				self.printer.Err(EMPTY_DICT_MSG)
			return None

		# Set that will be used to track which selected package names are
		# present in the result of the query
		presented_packages = set()

		# Convert a given iterable of tuples - each consisting of a single
		# package name - to a dict of package names
		def HandlePackages(packages):
			res = {}
			for name, group in packages:
				res[name] = \
					self.name_conv.Do(name, group, True, name_needed_params)
				presented_packages.add(name)
			return res
		# - nested func

		# A result is an iterable with a strictly defined order of elements
		# (it's easy to work on it in parallel) -- either a list of package
		# names or a matrix of package names
		if n_parts < 0:
			packages = HandlePackages(packages_cursor)
		else:
			packages = []
			# Splitting the data to a given number of parts
			data = SplitData(packages_cursor, n_parts)
			# Checking if all rows of a matrix of package names are empty
			if all([not section for section in data]):
				self.printer.Err(EMPTY_DICT_MSG)
				return None
			for section in data:
				part_of_packages = HandlePackages(section)
				if not part_of_packages:
					continue
				packages.append(part_of_packages)
		if not packages:
			if not selection:
				self.printer.Err(EMPTY_DICT_MSG)
				return None
			self.printer.Warn(EMPTY_DICT_MSG)
			return packages

		# Notifying about not present packages
		for name in selection:
			if name not in presented_packages:
				self.printer.LineAddWarn(f'There is no "{name}" package')

		self.printer.Success()

		return packages

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Get a dict of NVD/FSTEC software names with converted name and splitted
	# converted name that can be used for mapping and flag 'supplimentary' that
	# shows whether or not this name is listed in CPE dict (for NVD names only)

	def QueryProducts(self, type_of_products, selection=set()):

		if not self.__InitAuxObjects():
			return None

		self.printer.LineBegin(f'Getting a dict of {type_of_products.upper()} '
			'names')

		name_needed_params = NAME_NEEDED_PARAMS.get(self.type_of_matching)
		if not name_needed_params:
			self.printer.Err(
				f'Unknown type of matching "{self.type_of_matching}"')
			return None

		# Checking a name of a given type of products
		if type_of_products not in DATA_SOURCES:
			available_types = ' or '.join(DATA_SOURCES)
			self.printer.Err(f'Type of products should be {available_types}')
			return None

		# Connecting to a CVE database
		db = super().NewDB()
		if not db:
			return None

		# Checking an existence of a target table
		table = type_of_products + '_products'
		if not db.TableExists(table):
			self.printer.Err(f'Table "{table}" doesn\'t exist')
			return None

		is_multi_branch = self.type_of_matching in (M_URL, M_FIXES)

		# Getting product names
		query = "SELECT DISTINCT "
		if is_multi_branch:
			query += "vendor, product, group_id "
			query_group = ""
			products = defaultdict(dict)
		else:
			query += "'', product, GROUP_CONCAT(DISTINCT group_id SEPARATOR ' ') "
			query_group = "GROUP BY product "
			products = {}
		query += f"FROM {table} "

		conditions = []
		if selection:
			conditions.append(f'product in ({Placeholders(len(selection))})')
			data = [tuple(selection)]
		else:
			data = []
		if self.type_of_matching != M_URL:
			conditions.append('is_not_listed = 0')
		if conditions:
			query += 'WHERE ' + 'AND '.join(conditions) + ' '

		query += query_group

		products_cursor = db.Cursor(query, tables=(table,), data=data)
		if not products_cursor:
			if products_cursor != None:
				self.printer.Err(EMPTY_DICT_MSG)
			return None

		# Forming a dict of all products
		for vendor, product, groups in products_cursor:
			if is_multi_branch:
				# The "groups" var is a single group number in this case
				products[vendor][product] = \
					self.name_conv.Do(product, groups, False, name_needed_params)
			else:
				group = CombineGroups(map(int, groups.split()))
				products[product] = \
					self.name_conv.Do(product, group, False, name_needed_params)

		if not products:
			self.printer.Err(EMPTY_DICT_MSG)
			return None

		self.printer.Success()

		return products

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Send mathes of a particular type into a table with a given name

	def SendMatches(self, data_source, matches, field=None):

		if self.no_update:
			return True

		if not field:
			field = self.type_of_matching

		fields = [f'{field}_grade_{g}' for g in ('a', 'b')] if M_URL in field \
			else [field]

		table = f'{"bin" if self.target_is_bin else "src"}_to_{data_source}_map'

		self.printer.LineBegin(f'Updating "{", ".join(fields)}" '
			f'field{"s" if len(fields) > 1 else ""} of the "{table}" table')

		if not matches:
			self.printer.Warn('There is no matches')
			return True

		db = super().NewDB()
		if not db:
			return False

		query_to_clear = (
			f'UPDATE {table} SET ' +
				', '.join(f'{field}=NULL' for field in fields) + ' '
			'WHERE package=%s'
			)
		query_to_update = (
			f'INSERT INTO {table} (package, {", ".join(fields)}) '
			f'VALUES (%s, {Placeholders(len(fields))}) '
			'ON DUPLICATE KEY UPDATE ' +
				', '.join(f'{field}=%s' for field in fields)
			)
		data = [(package,) for package in matches.keys()]
		if db.Cursor(query_to_clear, tables=[table], data=data) == None:
			return False
		try:
			# With the executemany() method, it is not possible to specify
			# multiple statements to execute in the operation argument, so
			# it's not an option to use DB::Cursor because there is an INSERT
			# and an UPDATE operations in the formed query
			is_url_fields = all(M_URL in field for field in fields)
			for package, match in matches.items():
				data = (package,) + \
					tuple(match.get(f'grade_{g}') for g in ('a', 'b') * 2) \
					if is_url_fields else (package, match, match)
				db.cursor.execute(query_to_update, data)
			db.con.commit()
		except MySQLError as err:
			self.printer.Err(str(err))
			return False

		self.printer.Success()

		return True
