#!/usr/bin/python3

#	cve-manager : CVE management tool
#	Copyright (C) 2017-2022 Alexey Appolonov
#
#	This program is free software: you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation, either version 3 of the License, or
#	(at your option) any later version.
#
#	This program is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with this program.  If not, see <http://www.gnu.org/licenses/>.

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

from os                   import cpu_count
from types                import SimpleNamespace
from cve_manager.parallel import Parallel
from cpe_map.init         import Init
from cve_manager.conf     import DB_CON_SEC
from cpe_map.common       import NewArgParser
from cpe_map.control      import TerminateIfAbsent, TerminateOrSkip
from cpe_map.defines      import M_URL

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Prepare everything to perform a given type of matching (a type of matching
# that deals with branches, not just src_united or bin_united tables) and
# perform it by running a given function in parallel

def Run(type_of_matching, func_main, func_extra=[]):

	# Parsing the args and getting the helper objects
	args = NewArgParser(ptype='bp').parse_args()
	conf, speaker, mediator = Init(type_of_matching, args)

	# Getting a DB connection configuration
	mysql_config, = conf.Get([DB_CON_SEC])

	# Forming a list of branches that will be processed
	branches, err = \
		conf.SelectBranches(args.branches if 'all' not in args.branches else [])
	if err or not branches:
		speaker.Status(
			err=(err if err else 'Can\'t form a list of branch names'))
		return False

	# Getting package names
	mx_of_packages = mediator.QueryPackages(args.packages, args.names,
		cpu_count())
	TerminateIfAbsent(mx_of_packages)

	# Getting package URLs (if branch name is given) or product URLs (if not)
	def GetURLs(branch=None, data_source=None):
		msg = f"Processing URLs of {'packages' if branch else 'products'}"
		speaker.Say(msg)
		urls = func_extra[0](
			mysql_config,
			branch=branch,
			data_source=data_source,
			selection=(args.packages if branch else [])
			)
		if urls == None:
			speaker.Status(err="")
			return None, False
		if not urls:
			speaker.Status(warn='Empty dict, nothing to do')
			return None, True
		if branch:
			msg = f"Got URLs of {len(urls)} packages"
		else:
			msg = f"Got {sum(len(v) for v in urls.values())} URLs " \
				f"of {len(urls)} products"
		speaker.Status(success=msg)
		return urls, True
	# - nested func

	# For every type of selected data source (nvd, fstec)
	for data_source in args.data_sources:

		# Getting urls of products
		if type_of_matching == M_URL and func_extra:
			urls_of_products, ok = GetURLs(data_source=data_source)
			if not ok:
				return False
			elif not urls_of_products:
				return True
		else:
			urls_of_products = None

		# Getting product names
		products = mediator.QueryProducts(data_source)
		if TerminateOrSkip(products):
			continue

		for branch in branches:

			# Getting urls of packages
			if type_of_matching == M_URL and func_extra:
				urls_of_packages, ok = GetURLs(branch=branch)
				if not ok:
					return False
				elif not urls_of_packages:
					return True
			else:
				urls_of_packages = None

			# Gathering extra params for a given func to use
			setup = SimpleNamespace(
				mysql_config=mysql_config,
				data_source=data_source,
				branch=branch,
				products=products,
				urls_of_packages=urls_of_packages,
				urls_of_products=urls_of_products,
				)

			# Running multiple processes of matching
			speaker.Op(data_source, mx_of_packages, products, branch)
			matches, ok = Parallel(func_main, mx_of_packages, setup)
			if not ok:
				speaker.Status(err='Some process has terminated with an error')
				return False
			speaker.Status(matches=matches)

			# Updating the table with the results
			if not mediator.SendMatches(
					data_source, matches, f'{branch}_{type_of_matching}'):
				return False

	return True
