#!/usr/bin/python3

#	cve-manager : CVE management tool
#	Copyright (C) 2017-2022 Alexey Appolonov
#
#	This program is free software: you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation, either version 3 of the License, or
#	(at your option) any later version.
#
#	This program is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with this program.  If not, see <http://www.gnu.org/licenses/>.

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

import os
import re
from cve_manager.defines import M_PRESCRIBED, M_URL, M_COMPLETE, M_PARTIAL, \
	M_BIN_COMPLETE, M_BIN_PARTIAL, M_FIXES
from cpe_map.common      import FmtName
from cpe_map.groups      import G_PACKAGE_PREFIX, G_PRODUCT_PREFIX, \
	G_PRODUCT_SUFFIX

NAME_Q = 4
NAME_GROUP = 0
NAME_FORMATTED = 1
NAME_FORMATTED_SPLITTED = 2
NAME_FORMATTED_SPLITTED_WO_MARK = 3

NAME_NEEDED_PARAMS = {
	M_PRESCRIBED: {i: False for i in range(0, NAME_Q)},
	M_URL: {i: True if i == NAME_GROUP else False for i in range(0, NAME_Q)},
	M_COMPLETE: {
		NAME_GROUP: True,
		NAME_FORMATTED: True,
		NAME_FORMATTED_SPLITTED: False,
		NAME_FORMATTED_SPLITTED_WO_MARK: True,
		},
	M_PARTIAL: {
		NAME_GROUP: True,
		NAME_FORMATTED: False,
		NAME_FORMATTED_SPLITTED: True,
		NAME_FORMATTED_SPLITTED_WO_MARK: True,
		},
	}
NAME_NEEDED_PARAMS[M_FIXES]        = NAME_NEEDED_PARAMS[M_URL]
NAME_NEEDED_PARAMS[M_BIN_COMPLETE] = NAME_NEEDED_PARAMS[M_COMPLETE]
NAME_NEEDED_PARAMS[M_BIN_PARTIAL]  = NAME_NEEDED_PARAMS[M_PARTIAL]

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

class NameConverter():

	IS_PACKAGE_NAME = True
	IS_PRODUCT_NAME = not IS_PACKAGE_NAME
	CONV_BEGINING_STEPS = {}

	def __init__(self, conf_dir, groups):

		self.CONV_BEGINING_STEPS = {
			self.IS_PACKAGE_NAME:
				[(G_PACKAGE_PREFIX, self.__HandlePrefix)],
			self.IS_PRODUCT_NAME:
				[(G_PRODUCT_PREFIX, self.__HandlePrefix),
					(G_PRODUCT_SUFFIX, self.__HandleSuffix)],
				}

		self.groups = groups

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Search for a specified prefix/suffix in a specified name and cutout this
	# prefix/suffix

	def __HandlePrefix(self, name, prefix):

		found_prefix = re.search(fr'^{prefix}[^A-Za-z0-9]', name)
		if not found_prefix:
			return name, False

		return name[len(found_prefix.group(0)):], True


	def __HandleSuffix(self, name, suffix):

		found_suffix = re.search(fr'[^A-Za-z0-9]{suffix}$', name)
		if not found_suffix:
			return name, False

		return name[:-len(found_suffix.group(0))], True

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Represent given package name as a dictionary where:
	# * NAME_FORMATTED - formatted name (so that minor diffs won't prevent
	#   a match);
	# * NAME_FORMATTED_SPLITTED - list of words of formatted name (used during
	#   "partial" matching);
	# * NAME_FORMATTED_SPLITTED_WO_MARK - list of words of formatted name
	#   without special prefix and special suffix if they were found, or None
	#   value if they were not found

	def Do(self, name, group, is_package_name, needed_params,
			highlight_lib=False):

		# Format given name (symbols can be replaced but not removed, except
		# delimiter symbols at the end, that will be removed)
		name = FmtName(name)

		sname = name
		got_spec_prefix_or_suffix = False

		if name and needed_params[NAME_FORMATTED_SPLITTED_WO_MARK]:

			params_of_this_group = self.groups.get(group, {})

			# Removing common prefixes and suffixes
			if params_of_this_group:
				for param, f in self.CONV_BEGINING_STEPS[is_package_name]:
					for prefix_or_suffix in params_of_this_group.get(param, []):
						conv_name, got_spec_prefix_or_suffix = \
							f(sname, FmtName(prefix_or_suffix))
						if got_spec_prefix_or_suffix:
							sname = conv_name
							break
					if got_spec_prefix_or_suffix:
						break

			# Highlighting the "lib" prefix/suffix
			if highlight_lib:
				sname = \
					re.sub(r'^lib(?!_)', 'lib_',
					re.sub(r'(?<!_)lib$', '_lib',
					name))

		res = {}
		for i in range (0, NAME_Q):
			if not needed_params[i]:
				continue
			if i == NAME_GROUP:
				res[i] = group
			elif i == NAME_FORMATTED:
				res[i] = name
			elif i == NAME_FORMATTED_SPLITTED:
				res[i] = name.split('_')
			elif i == NAME_FORMATTED_SPLITTED_WO_MARK:
				res[i] = sname.split('_') if got_spec_prefix_or_suffix else None

		return res
