#!/usr/bin/python3

#	cve-manager : CVE management tool
#	Copyright (C) 2017-2022 Alexey Appolonov
#
#	This program is free software: you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation, either version 3 of the License, or
#	(at your option) any later version.
#
#	This program is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with this program.  If not, see <http://www.gnu.org/licenses/>.

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

import gc
import pickle
from os   import path
from copy import deepcopy

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

class Worker():

	def __init__(self, home_dir):

		self.__ILIST = []           # Indexes list-of-lists
		self.ILIST_ROWS = 0         # Number of rows of indexes list of lists
		self.ILIST_COLS = 0         # Number of cols of indexes list oflists
		self.__ILIST_ROWS_LIM = 7   # Max limit of rows of the 'ilist' matrix
		self.__ILIST_COLS_LIM = 7   # Max limit of cols of the 'ilist' matrix
		self.MIN_SCORE = 0.8        # Min score that a matching could have
		self.MIN_ACCEPTABLE_LEN = 7 # Min length of names to perform partial matching
		self.PREFIX_PENALTY = 0.95  # Penalty for having a special prefix
		self.__ILIST_PATH = path.join(home_dir, '.ilist')
		                            # Path for the list of matrixes of indexes

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Forming ILIST and saving it to the disk

	def __SaveIList(self):

		self.__ILIST = self.__FormIList(self.ILIST_ROWS, self.ILIST_COLS)

		with open(self.__ILIST_PATH, 'wb') as f:
			pickle.dump(self.__ILIST, f)

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Loading ILIST from the disk and reforming it if it's size isn't sufficient

	def __LoadIList(self):

		if any((d.get('rows') == self.ILIST_ROWS and \
				d.get('cols') == self.ILIST_COLS) for d in self.__ILIST):
			return

		with open(self.__ILIST_PATH, 'rb') as f:
			self.__ILIST = pickle.load(f)

		if not self.__CheckIListSize():
			self.__SaveIList()

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Forming [n x m] matrix of indexes

	def __FormMatrixOfIndexes(self, n, m):

		if n < 1 or m < 1:
			return []

		matrix = []

		for i in range(0, n):
			vector = []
			for j in range(0, m):
				vector.append((i, j))
			matrix.append(vector)

		return matrix

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Forming ILIST by calling FormMatrixOfIndexes for every
	#	{rows, cols} combination up to [n x m]

	def __FormIList(self, n, m):

		if n < 1 or m < 1:
			return []

		ilist = []

		for rows in range(1, n + 1):
			for cols in range(1, m + 1):
				list_of_lists_of_indexes = \
					self.GetListOfListsOfIndexes(rows, cols, ilist)
				if list_of_lists_of_indexes:
					continue
				list_of_lists_of_indexes = \
					self.GetListOfListsOfIndexes(cols, rows, ilist)
				if list_of_lists_of_indexes:
					ilist.append({'rows': rows, 'cols': cols,
						'list': self.__SwapRowsAndCols(list_of_lists_of_indexes)})
				else:
					if cols > rows:
						x = cols
						y = rows
					else:
						x = rows
						y = cols
					matrix = self.__FormMatrixOfIndexes(x, y)
					list_of_lists_of_indexes = \
						self.__SubMatrix(matrix, self.__FormMask(matrix))
					ilist.append({'rows': x, 'cols': y,
						'list': list_of_lists_of_indexes})
					if x != y and y <= n and x <= m:
						ilist.append({'rows': y, 'cols': x,
							'list': self.__SwapRowsAndCols(list_of_lists_of_indexes)})
				gc.collect()

		return ilist

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Forming rows/cols mask used in SubMatrix algorithm

	def __FormMask(self, matrix):
		return {
			'rows': list(range(0, len(matrix))),
			'cols': list(range(0, len(matrix[0])))
			}

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Checking whether ILIST actual size matches ILIST_ROWS & ILIST_COLS values

	def __CheckIListSize(self):

		max_rows = 0
		max_cols = 0

		for el in self.__ILIST:
			if el['rows'] > max_rows:
				max_rows = el['rows']
			if el['cols'] > max_cols:
				max_cols = el['cols']

		if max_cols < self.ILIST_COLS or max_rows < self.ILIST_ROWS:
			return False

		return True

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Forming list of combinations out of matrix of combinations

	def __SubMatrix(self, matrix, mask):

		if len(mask['rows']) < 1:
			return [[]]

		if len(mask['cols']) <= 1:
			return self.__GetColumn(
				matrix[mask['rows'][0]:(mask['rows'][-1] + 1)],
				mask['cols'][0]
				)

		list_of_combinations = []
		sub_mask = deepcopy(mask)
		del sub_mask['rows'][0]

		for j in mask['cols']:
			sub_mask['cols'].remove(j)
			sub_matrix = self.__SubMatrix(matrix, sub_mask)
			sub_mask['cols'].append(j)
			new_list_of_combinations = \
				self.__AppendToEach(sub_matrix, matrix[mask['rows'][0]][j])
			list_of_combinations += new_list_of_combinations

		return list_of_combinations

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Getting column from matrix

	def __GetColumn(self, matrix, n_column):
		return [[row[n_column]] for row in matrix ]

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Appending element to every list in the list of lists

	def __AppendToEach(self, list_of_lists, new_element):

		for l in list_of_lists:
			l.append(new_element)

		return list_of_lists

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Swapped matrix of indexes for [n x m] combinations matrix
	#	equals matrix of index for [m x n] combinations matrix

	def __SwapRowsAndCols(self, list_of_lists_of_tuples):

		matrix = []

		for row in list_of_lists_of_tuples:
			vector = []
			for el in row:
				vector.append((el[1], el[0]))
			matrix.append(vector)

		return matrix

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Getting a matrix of indexes  with [rows x cols] size from ILIST

	def GetListOfListsOfIndexes(self, rows, cols, ilist=None):

		if ilist == None:
			ilist = self.__ILIST

		for d in ilist:
			if d['rows'] == rows and d['cols'] == cols:
				return d['list']

		return []

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Preparing matrix of indexes

	def Prepare(self, rows, cols, force):

		self.ILIST_ROWS = min(rows, self.__ILIST_ROWS_LIM)
		self.ILIST_COLS = min(cols, self.__ILIST_COLS_LIM)

		if force or not path.isfile(self.__ILIST_PATH):
			self.__SaveIList()
		else:
			self.__LoadIList()

		return self.__ILIST and self.ILIST_ROWS > 0 and self.ILIST_COLS > 0
