#!/usr/bin/python3

#	cve-manager : CVE management tool
#	Copyright (C) 2017-2022 Alexey Appolonov
#
#	This program is free software: you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation, either version 3 of the License, or
#	(at your option) any later version.
#
#	This program is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with this program.  If not, see <http://www.gnu.org/licenses/>.

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

from mysql.connector  import Error as MySQLError
from collections      import defaultdict
from cve_manager.conf import LOCAL_SYS
from cve_manager.db   import DB
from cpe_map.defines  import NAME_MAX_LEN

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

class Worker():

	def __init__(self, mysql_config, printer, branches):

		self.mysql_config = mysql_config
		self.printer = printer
		self.branches = branches

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Check existence of src_united and bin_united tables

	def __CheckUnitedTables(self, recreate):

		src_united = 'src_united'
		bin_united = 'bin_united'

		self.printer.LineBegin(f'Checking {src_united} and {bin_united} tables')

		db = DB(self.printer)
		if not db.Connect(self.mysql_config):
			return None, False

		src_united_exists = db.TableExists(src_united)
		if src_united_exists == None:
			return None, False

		bin_united_exists = db.TableExists(bin_united)
		if bin_united_exists == None:
			return None, False

		both_exist = src_united_exists and bin_united_exists

		# Forming a status message
		msg = ''
		skipping = False
		if both_exist:
			if recreate:
				msg = 'Both exist, request to recreate'
			else:
				skipping = True
		else:
			if src_united_exists or bin_united_exists:
				if src_united_exists:
					msg = f'{src_united} exists and {bin_united} does not'
				else:
					msg = f'{bin_united} exists and {src_united} does not'
				msg += ', should recreate'
			else:
				msg = 'Both not exist, should create'

		if skipping:
			self.printer.LineEnd('[SKIPPING: Already exist]')
		else:
			self.printer.Success(msg)

		return both_exist, True

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Get a set of src/bin package names

	def __QueryPackages(self, branch, target):

		msg = f'Getting a set of {target} package names of ' + \
			('the local system' if branch == LOCAL_SYS else f'{branch} branch')
		self.printer.LineBegin(msg)

		db = DB(self.printer)
		if not db.Connect(self.mysql_config):
			return {}, False

		if target == 'src':
			tables = [branch + '_src']
			query  = f"SELECT name, '' FROM {tables[0]}"
		else:
			tables = [branch + '_bin' , branch + '_src']
			query  = f'SELECT bin.name, src.name FROM {tables[0]} bin ' \
				f'INNER JOIN {tables[1]} src ON src.id = bin.src_id ' \
				f'WHERE bin.name <> src.name'

		if not db.Cursor(query, tables=tables):
			return {}, False

		packages = defaultdict(set)
		for a, b in db.cursor:
			s = packages[a]
			if b:
				s.add(b)

		return packages, self.printer.BiStatus(packages)

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Send united package names to the DB

	def __SendUnitedPacks(self, names_and_groups_of_packages, target):

		self.printer.LineBegin(f'Sending {target} united package names')

		if target not in ('src', 'bin'):
			self.printer.Err(f'Wrong target "{target}"')
			return False

		if not names_and_groups_of_packages:
			self.printer.Err(f'Empty list of {target} packages')
			return False

		db = DB(self.printer)
		if not db.Connect(self.mysql_config):
			return False

		# Creating the table and inserting the data
		table = f'{target}_united'
		data = names_and_groups_of_packages
		columns = [
			f'name VARCHAR({NAME_MAX_LEN}) COLLATE utf8mb4_0900_as_cs',
			'group_id TINYINT',
			]
		query = f'INSERT INTO {table} (name, group_id) VALUES (%s, %s)'
		if not db.RecreateTable(table, columns) or \
				db.Cursor(query, tables=[table], data=data) == None:
			return False

		self.printer.Success()

		return True

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Prepare tables for united package names

	def Prepare(self, groups, recreate):

		# Don't do enything if src_united and bin_united tables already exist
		# and there is no request to recreate them
		united_tables_exists, ok = self.__CheckUnitedTables(recreate)
		if not ok:
			return False
		if united_tables_exists and not recreate:
			return True

		for t in ('src', 'bin'):

			is_src_target = (t == 'src')
			packages_united = defaultdict(set)
			names_and_groups_of_packages = []

			for branch in self.branches:
				packages, ok = self.__QueryPackages(branch, t)
				if not ok:
					return False
				for src_or_bin_name, src_names in packages.items():
					packages_united[src_or_bin_name] |= src_names

			for src_or_bin_name, src_names in packages_united.items():
				if is_src_target:
					group = groups.CheckPackage(src_or_bin_name)
				else:
					# Group of a bin package...
					# ...is 0 if groups of all of it's src packages are 0
					# ...is -1 if groups of at least two of it's src packages
					# are greater than 0 and not equal to each other
					# ...is a group of it's src packages that is not 0, in all
					# other cases
					group = 0
					for src_name in src_names:
						g = groups.CheckPackage(src_name)
						if group == 0:
							group = g
							continue
						if g != group:
							group = -1
							break
				names_and_groups_of_packages.append((src_or_bin_name, group))

			if not self.__SendUnitedPacks(names_and_groups_of_packages, t):
				return False

		return True
