#!/usr/bin/python3

#	cve-manager : CVE management tool
#	Copyright (C) 2017-2022 Alexey Appolonov
#
#	This program is free software: you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation, either version 3 of the License, or
#	(at your option) any later version.
#
#	This program is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with this program.  If not, see <http://www.gnu.org/licenses/>.

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

import re
from mysql.connector       import Error as MySQLError
from ax.ver                import GetSingleVerAndRLeftRRight
from collections           import defaultdict
from cve_manager.common    import UntaggedRefs
from cve_manager.db        import DB, Placeholders
from cve_manager.defines   import DATA_SOURCES, NVD_DATA_SRC, FSTEC_DATA_SRC
from cve_manager.software  import DistinctSoftware
from cve_manager.url       import ParseURL
from cpe_map.defines       import IGNORED_NETLOCS, NAME_MAX_LEN
from cpe_map.cpe_relatives import PrescribedCPERelatives

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Format given product name to a (root, prefix_id, suffix_id) tuple

NON_LETTERS = '[0-9\-\.,_]'
SPEC_MARKS  = {
	'lib': 1,
	}

def FmtProductName(name):

	root = re.sub(fr'(^{NON_LETTERS}+|{NON_LETTERS}+$)', '', name).lower()
	prefix_id = 0
	suffix_id = 0

	for k, v in SPEC_MARKS.items():
		if not prefix_id and root.startswith(k):
			prefix_id = v
			root = root[len(k):]
		if not suffix_id and root.endswith(k):
			suffix_id = v
			root = root[:-len(k)]

	root = re.sub(r'[\-\.,_]', '', root)

	return root, prefix_id, suffix_id

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Check that specified product names are related

def AreRelatedProductNames(a, b):

	x = FmtProductName(a)
	y = FmtProductName(b)

	if x[0] != y[0]:
		return False

	for i in (1, 2):
		if x[i] and y[i] and x[i] != y[i]:
			return False

	for p, q in ((x, y), (y, x)):
		if p[1] and not q[1]:
			if q[2]:
				if p[1] != q[2]:
					return False
			else:
				if p[2]:
					return False

	return True

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

class Worker():

	def __init__(self, conf_dir, mysql_config, printer, types_of_products):

		self.conf_dir = conf_dir
		self.db = None
		self.mysql_config = mysql_config
		self.printer = printer
		self.types_of_products = types_of_products
		self.products_table  = ''
		self.timelines_table = ''
		self.vul_table = ''
		self.cpe_table = 'cpe_import'

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Initialize product tables considering given type of vul lists

	def __InitTableNames(self, type_of_products):

		self.products_table  = type_of_products + '_products'
		self.timelines_table = self.products_table + '_timelines'
		self.vul_table = type_of_products + '_vul_import'
		self.current_type_of_products_is_fstec = \
			type_of_products == FSTEC_DATA_SRC

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Create a *_products or a *_products_timelines table

	def __CreateProductsTable(self, table, recreate):

		self.printer.LineBegin(f'Creating "{table}" table')

		if table == self.products_table:
			columns = [
				'id INT UNSIGNED UNIQUE',
				f'vendor VARCHAR({NAME_MAX_LEN})',
				f'product VARCHAR({NAME_MAX_LEN})',
				'is_supplimentary BOOLEAN DEFAULT NULL',
				'is_non_free BOOLEAN DEFAULT NULL',
				'is_not_listed BOOLEAN DEFAULT NULL',
				'relatives TEXT',
				'refs LONGTEXT',
				'group_id TINYINT',
				f'CONSTRAINT pk_{table} PRIMARY KEY (id)',
				]
		elif table == self.timelines_table:
			columns = [
				'id INT UNSIGNED',
				'ver VARCHAR(64)',
				'first_mention DATE',
				f'CONSTRAINT pk_{table} PRIMARY KEY (id, ver)',
				f'CONSTRAINT fk_{table} FOREIGN KEY (id) '
					f'REFERENCES {self.products_table} (id)',
				]
		else:
			self.printer.Err('Unknown table')
			return -1

		if self.db.TableExists(table) and not recreate:
			self.printer.LineEnd('[SKIPPING: Already exist]')
			return 0

		try:
			if table == self.products_table:
				self.db.DropTable(self.timelines_table)
			self.db.RecreateTable(table, columns)
		except MySQLError as err:
			self.printer.Err(f'{err}')
			return -1

		self.printer.Success()

		return 1


	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Check existence of tables that are source tables for a *_products table

	def __CheckRequiredTables(self, type_of_products, just_created):

		self.printer.LineBegin(f'Checking the existence of the required tables')

		tables_check = {}

		if just_created[self.products_table] and \
				just_created[self.timelines_table]:
			tables_check[self.vul_table] = self.db.TableExists(self.vul_table)

		if just_created[self.products_table] and \
				type_of_products == NVD_DATA_SRC:
			tables_check[self.cpe_table] = self.db.TableExists(self.cpe_table)

		for table, ok in tables_check.items():
			if not ok:
				self.printer.Err(f'Table "{table}" doesn\'t exist')
				return False

		self.printer.Success()

		return True

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Get all products from CPE dict

	def __GetCPEDict(self):

		self.printer.LineBegin('Getting all products from CPE dict')

		cpe_dict = defaultdict(dict)
		cols = ['vendor', 'product', 'refs', 'non_free']

		if not self.db.Select(cols, self.cpe_table):
			self.printer.Err(f"Can't get CPEs from {self.cpe_table} table")
			return None

		for vendor, product, refs, non_free in self.db.cursor:
			# Value of non_free is 2 if all titles of this product state
			# that this app is for non-free OS
			cpe_dict[product][vendor] = (refs, non_free == 2)

		self.printer.Success()

		return cpe_dict

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Search for related product names and modify the <relatives> element of
	# the products_info dict of dict of tuples

	def __IdentifyRelatives(self, products_info, products_refs):

		# Add product "b" to relatives of product "a" and vice versa
		# (both arguments are tuples (<vendor>, <product>))
		def AddRelatives(a, b):
			products_info_a = products_info.get(a[1], {}).get(a[0])
			products_info_b = products_info.get(b[1], {}).get(b[0])
			if products_info_a and products_info_b:
				products_info_a[1].add(b)
				products_info_b[1].add(a)
		# - nested func

		# Processing products specified in a same list
		for vendor_product_pairs in products_refs:
			for i, vendor_and_product in enumerate(vendor_product_pairs[:-1]):
				vendor_a, product_a = vendor_and_product
				for vendor_b, product_b in vendor_product_pairs[i + 1:]:
					if AreRelatedProductNames(product_a, product_b):
						AddRelatives((vendor_a, product_a), (vendor_b, product_b))

		products = defaultdict(list)

		# Identifying products that have same vendors
		for product, vendors in products_info.items():
			for vendor in vendors:
				products[vendor].append(product)

		# Processing products that have same vendors
		for vendor, product_names in products.items():
			for i, product_a in enumerate(product_names[:-1]):
				for product_b in product_names[i + 1:]:
					if AreRelatedProductNames(product_a, product_b):
						AddRelatives((vendor, product_a), (vendor, product_b))

		# Processing products that have vendors that differ only in the
		# presence or absence of the "_project" suffix
		for vendor_a, product_names_a in products.items():
			vendor_b = vendor_a + '_project'
			for product_b in products.get(vendor_b, []):
				for product_a in product_names_a:
					if AreRelatedProductNames(product_a, product_b):
						AddRelatives((vendor_a, product_a), (vendor_b, product_b))

		prescriptions, warn = PrescribedCPERelatives(self.conf_dir).Read()

		# Add related products from the list of prescribed CPE relatives
		for product, related_products in prescriptions.items():
			for related_product in related_products:
				AddRelatives(product, related_product)

		return warn

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Remove entries with too long product names from a dict of products info
	# that is a dict of dicts {<product>: {<vendor>: {...}}}

	def __FilterProductsInfo(self, products_info):

		filtered_products_info = {}
		waste = {}

		for product, vendors in products_info.items():
			if len(product) > NAME_MAX_LEN:
				waste[product] = vendors.keys()
				continue
			filtered_products_info[product] = vendors

		return (filtered_products_info, waste)

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Get products info from the <type_of_products>_vul_import table

	def __GetProductsInfo(self, type_of_products, cpe_dict):

		self.printer.LineBegin('Getting products info from the '
			f'{type_of_products}_vul_import table')

		table = self.vul_table
		cols  = ['software_list', 'refs'] + \
			(["''", 'published'] if type_of_products == FSTEC_DATA_SRC else
				['patch', 'modified'])
		cond = "software_list <> ''"

		if not self.db.Select(cols, table, cond):
			self.printer.Err('There are no rows with filled software_list '
				f'column in {table} table')
			return (None,) * 3

		# {<product> :
		#     {<vendor> :
		#         (<id>, <relatives>, <refs>, <group>, {<ver> : <first_mention>})
		#         }
		#     }
		products_info = defaultdict(lambda: defaultdict(
			lambda: [0, set(), set(), 0, {}]))

		products_refs = defaultdict(set)
		product_counter = 1

		# Selecting all the needed data
		for software_listing, refs, patch, modified in self.db.cursor:
			products_of_this_software_list = set()
			untagged_filtered_refs = set()
			for ref in (UntaggedRefs(refs).split() + patch.split()):
				url = ParseURL(ref)
				if any(re.search(fr'(^|\.){netloc}$', url["netloc"])
						for netloc in IGNORED_NETLOCS):
					continue
				untagged_filtered_refs.add(ref)
			for software in DistinctSoftware(software_listing):
				vendor = software.Vendor()
				product = software.Product()
				if self.current_type_of_products_is_fstec:
					vendor = vendor.lower()
					product = product.lower()
				products_of_this_software_list.add((vendor, product))
				if vendor and product:
					ver_or_range_of_vers = software.Ver()
					single_ver, _ , range_right = \
						GetSingleVerAndRLeftRRight(ver_or_range_of_vers)
					ver = single_ver if single_ver else range_right
					if ver and self.current_type_of_products_is_fstec:
						ver = ver.lower()
					product_id, _, _, _, product_timeline = products_info \
						.get(product, {}) \
						.get(vendor, [0, set(), set(), 0, {}])
					# If there is no entry for this product
					if not product_id:
						products_info[product][vendor] = \
							[product_counter, set(), set(), 0, product_timeline]
						product_counter += 1
						modified_0 = None
					else:
						modified_0 = product_timeline.get(ver)
					# If date of existing entry is earlier
					if modified_0 and modified_0 <= modified:
						continue
					product_timeline[ver] = modified
			products_of_this_software_list = \
				tuple(sorted(products_of_this_software_list))
			products_refs[products_of_this_software_list] |= \
				untagged_filtered_refs

		warn = self.__IdentifyRelatives(products_info, products_refs)
		if warn:
			self.printer.LineAddWarn(warn)

		# Aggregating the products refferences
		for vendor_product_pairs, refs in products_refs.items():
			# Check that exactly one product name is specified in the software list
			single_product = \
				len({product for _, product in vendor_product_pairs}) == 1
			# Check that all products of the software list are relatives
			only_relatives = None if single_product else all(
				vendor_and_product in products_info[product][vendor][1]
				for i, vendor_and_product in enumerate(vendor_product_pairs)
				for vendor, product in vendor_product_pairs[i + 1:]
				)
			# Addings the refferences if there is only one product
			# or if all products are related
			if single_product or only_relatives:
				for vendor, product in vendor_product_pairs:
					products_info[product][vendor][2] |= refs

		products_not_listed = defaultdict(set)

		# Adding products that are not mentioned in the vulnerability lists but
		# whose names are the same as names of some of the mentioned products
		for product, vendors in cpe_dict.items():
			mentioned_vendors = products_info.get(product)
			if not mentioned_vendors:
				continue
			for vendor in vendors:
				if vendor not in mentioned_vendors:
					products_info[product][vendor] = [product_counter, set(), set(), 0, {}]
					products_not_listed[product].add(vendor)
					product_counter += 1

		if not products_info:
			msg = 'There are no products'
			if type_of_products == FSTEC_DATA_SRC:
				self.printer.Warn(msg)
			else:
				self.printer.Err(msg)
			return (None,) * 3

		self.printer.Success()

		# Removing products with too long names
		products_info, waste = self.__FilterProductsInfo(products_info)

		return products_info, products_not_listed, waste

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Identify groups of products using the "groups" object and write it to the
	# "products_info[<product>][<vendor>][3]" list element

	def __IdentifyGroupsOfProducts(self, products_info, groups, cpe_dict):

		if not groups or not cpe_dict:
			return

		for product, vendors in products_info.items():
			for vendor, info in vendors.items():
				refs_str, _ = cpe_dict.get(product, {}).get(vendor, ('', None))
				refs = info[2] | {ref for ref in refs_str.split() if ref}
				info[3] = groups.CheckProduct(vendor, product, refs)

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Send data to the *_products table; Some of the fields are:
	# is_supplimentary
	#   -- True if this product is not present in the CPE dict;
	# is_non_free
	#   -- True if all titles of this product in the CPE dict state that this
	#   app is for non-free OS;
	# is_not_listed
	#   -- True if this product is not mentioned in the vulnerability lists
	#   but whose name is the same as names of some of the mentioned products;

	def __FillProductsTable(self, products, products_not_listed, waste,
			cpe_dict):

		self.printer.LineBegin(f'Filling "{self.products_table}" table')

		query = f'INSERT INTO {self.products_table} (id, vendor, product, ' \
			'is_supplimentary, is_non_free, is_not_listed, relatives, refs, ' \
			f'group_id) VALUES ({Placeholders(9)})'
		data = []

		for product, vendors in products.items():
			for vendor, info in vendors.items():
				product_id, relatives, untagged_filtered_refs, group, _ = info
				if self.current_type_of_products_is_fstec:
					is_supplimentary = False
					is_non_free = False
					is_not_listed = False
				else:
					_, cpe_flag = cpe_dict \
						.get(product, {})  \
						.get(vendor, (None, None))
					is_supplimentary = cpe_flag == None
					is_non_free = cpe_flag if cpe_flag != None else False
					is_not_listed = vendor in products_not_listed.get(product, {})
				if is_non_free and is_not_listed:
					continue
				joined_relatives = ' '.join([f'{vendor}:{product}'
					for vendor, product in sorted(relatives)])
				joined_untagged_filtered_refs = ' '.join(untagged_filtered_refs)
				data.append((
					product_id,
					vendor,
					product,
					is_supplimentary,
					is_non_free,
					is_not_listed,
					joined_relatives,
					joined_untagged_filtered_refs,
					group,
					))

		if not data:
			self.printer.Err('Empty list of product names')
			return False

		if not self.db.Cursor(query, data=data):
			return False

		if waste:
			for product in sorted(waste):
				for vendor in sorted(waste[product]):
					self.printer.LineAddExtra(f'{vendor}:{product}')
			self.printer.Warn('Some of the product names are too long '
				'and were dropped')
		else:
			self.printer.Success(f'{len(data)} names total')

		return True

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Send data to the timelines table

	def __FillTimelinesTable(self, products):

		self.printer.LineBegin(f'Filling "{self.timelines_table}" table')

		query = f'INSERT INTO {self.timelines_table} ' \
			'(id, ver, first_mention) VALUES (%s, %s, %s)'
		data = []

		for product, vendors in products.items():
			for vendor, info in vendors.items():
				product_id, _, _, _, product_timeline = info
				for ver, first_mention in product_timeline.items():
					if not ver or not first_mention \
							or ver in ('_nil_', '-', '.'):
						continue
					data.append((product_id, ver, first_mention))

		if not data:
			self.printer.Err('No timelines of products')
			return False

		if not self.db.Cursor(query, data=data):
			return False

		self.printer.Success()

		return True

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Create a table that will hold all NVD\FSTEC product names

	def __CreateAndFillProductTables(self, type_of_products, groups, recreate):

		self.__InitTableNames(type_of_products)

		# Creating a table for products and a table for timelines if not exist
		# or in case of a special request
		just_created = {}
		for table in (self.products_table, self.timelines_table):
			status = self.__CreateProductsTable(table, recreate)
			if status < 0:
				return False
			just_created[table] = (status == 1)

		# Leaving if both tables already existed
		if all(not v for v in just_created.values()):
			return True

		# Checking the existence of the required tables
		if not self.__CheckRequiredTables(type_of_products, just_created):
			return False

		# Getting all products from CPE dict
		cpe_dict = self.__GetCPEDict() if type_of_products == NVD_DATA_SRC \
			and just_created[self.products_table] else {}
		if cpe_dict == None:
			return False

		# Getting products info from the *_vul_import table
		products, products_not_listed, waste = \
			self.__GetProductsInfo(type_of_products, cpe_dict)
		if any(el == None for el in (products, products_not_listed, waste)):
			return False

		# Identifying groups that the products belong to
		self.__IdentifyGroupsOfProducts(products, groups, cpe_dict)

		# Filling a *_products table
		if just_created[self.products_table] and \
				not self.__FillProductsTable(
					products, products_not_listed, waste, cpe_dict):
			return False

		# Filling a timelines table
		if just_created[self.timelines_table] and \
				not self.__FillTimelinesTable(products):
			return False

		return True

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Create and fill tables for (united) product names and their timelines

	def Prepare(self, groups, recreate):

		# Connecting to the database
		self.db = DB(self.printer)
		if not self.db.Connect(self.mysql_config):
			return False

		for type_of_products in self.types_of_products:

			# Validating a given type of vul lists
			if type_of_products not in DATA_SOURCES:
				self.printer.Err(f'Can\'t create "{products_table}" table - '
					'wrong type of products, type of products should be ' +
					' or '.join(DATA_SOURCES))
				return False

			if not self.__CreateAndFillProductTables(type_of_products, groups, recreate) \
					and type_of_products != FSTEC_DATA_SRC:
				return False

		return self.db.Disconnect()
