#!/usr/bin/python3

#	cve-manager : CVE management tool
#	Copyright (C) 2017-2022 Alexey Appolonov
#
#	This program is free software: you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation, either version 3 of the License, or
#	(at your option) any later version.
#
#	This program is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with this program.  If not, see <http://www.gnu.org/licenses/>.

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

from re              import search
from collections     import defaultdict
from datetime        import date
from ax.ver          import IsDateVer, CompareSVersions
from cve_manager.url import ParseAndJoinURL
from cpe_map.defines import M_PRESCRIBED
from cpe_map.common  import FormCPEStyleStr

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

class Filter():

	def __init__(self, ignore, current_versions_of_packages,
			time_lines_of_packages, time_lines_of_products):

		self.ignore = ignore
		self.current_versions_of_packages = current_versions_of_packages
		self.time_lines_of_packages = time_lines_of_packages
		self.time_lines_of_products = time_lines_of_products

		# Flattening a dict of timelines of products into a set
		buf = defaultdict(dict)
		for products_of_vendor in self.time_lines_of_products.values():
			for product, info in products_of_vendor.items():
				for ver, first_mention in info:
					first_mention0 = buf[product].get(ver)
					if not first_mention0 or first_mention <= first_mention0:
						buf[product][ver] = first_mention
		self.time_lines_of_products_flat = defaultdict(list)
		for product, first_mentions in buf.items():
			for ver, first_mention in first_mentions.items():
				self.time_lines_of_products_flat[product]\
					.append((ver, first_mention))

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Check that a match between specified package and product should be ignored

	def __NotIgnored(self, branch, package_name, package_url, vendor, product):


		if self.ignore:

			excl = self.ignore.get('', {}).get(package_name, {}) | \
				self.ignore.get(branch, {}).get(package_name, {})

			# URLs of packages of excluded pairs are processed with the
			# "ParseAndJoinURL" function...
			package_url = ParseAndJoinURL(package_url)

			for iurl, iproducts in excl.items():
				if not iurl or (iurl.startswith(package_url)
						if len(iurl) >= len(package_url)
						else package_url.startswith(iurl)):
					for ivendor, iproduct in iproducts:
						if ivendor in ('', '*', vendor) and \
								iproduct in ('*', product):
							cpe = FormCPEStyleStr(vendor, product)
							msg = f'Ignoring "{package_name} >> {cpe}" choice'
							return False, msg

		return True, ''

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Check that timelines of specified package and product do match

	def __TimelinesMatch(self, branch, package_name, vendor, product,
			package_relatives):

		# It's OK if there is no timeline of a package
		package_timeline = self.time_lines_of_packages.get(package_name, [])
		if not package_timeline:
			msg = f'Can\'t get a time line of the {package_name} package'
			return True, msg

		got_mismatch = False
		product_timeline = (
			self.time_lines_of_products_flat.get(product, [])
			if vendor in ('', '*') else
			self.time_lines_of_products.get(vendor, {}).get(product, [])
			)

		# All first mentions of versions of a product should be latter
		# than timestamps of same or bigger versions of a package
		for package_ver, package_timestamp in package_timeline:
			package_date = date.fromtimestamp(package_timestamp)
			package_ver_is_date = IsDateVer(package_ver)
			for product_ver, product_date in product_timeline:
				# If this version of a this package was build before
				# this version of this product was found vulnerable
				if package_date < product_date:
					continue
				# Not comparing normal versions with date-versions
				if package_ver_is_date != IsDateVer(product_ver):
					continue
				comp, symbolic = CompareSVersions(package_ver, product_ver)
				# Symbolic versions do not count
				if symbolic:
					continue
				# If there is a same or lesser version of this product that was
				# mentioned before this version of this package was build
				# than this pair should not be recognized as mismatch
				if comp >= 0:
					return True, ''
				# Forming only one message for every mismatched pair
				if not got_mismatch:
					cpe = FormCPEStyleStr(vendor, product)
					msg = 'Dates mismatch in "' \
						f'{package_name} ({package_ver}, {package_date}) >> ' \
						f'{cpe} ({product_ver}, {product_date})" choice'
					got_mismatch = True

		if got_mismatch:
			# All the relatives should fail too
			for rpackage_name, rpackage_url in package_relatives:
				ok, _ = self.Pass(branch, rpackage_name, rpackage_url, vendor,
					product)
				if ok:
					return True, ''
			return False, msg

		return True, ''

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Check that a package version and a version that can be included in a
	# package name don't conflict with a version that can be included in a
	# product name (a package version isn't greater than a product version)

	def __VersionsInNamesMatch(self, package_name, product_name):

		versions = []

		for name in (package_name, product_name):
			s = search(r'[0-9][0-9,\.\_\-]*$', name)
			if not s:
				return True, ''
			versions.append(s.group())

		if CompareSVersions(*versions)[0] <= 0:
			return True, ''

		package_ver = self.current_versions_of_packages.get(package_name)
		product_ver = versions[1]

		if not package_ver:
			msg = f'Can\'t find version of a "{package_name}" package'
			return False, msg

		if CompareSVersions(package_ver, product_ver)[0] <= 0:
			return True, ''

		msg = f'Version of a "{package_name}" package is greater ' \
			f'than version stated in a product name "{product_name}"'
		return False, msg

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Check a given pair of package and product and return a boolean value and
	# a clarifying message string (optional); The boolean value is True if:
	# 1) that pair should not be ignored and
	# 2) there is no mismatch between timelines of the package and the product;
	# The boolean value is False otherwise

	def Pass(self, branch, package_name, package_url, vendor, product,
			package_relatives=[], type_of_matching=''):

		# Not performing any checks for prescribed pairs
		if type_of_matching == M_PRESCRIBED:
			return True, ''

		# Considering an 'ignore' dict
		ok, msg = self.__NotIgnored(branch, package_name, package_url, vendor,
			product)
		if not ok:
			return False, msg

		# Considering timelines of a packages and a product
		ok, msg = self.__TimelinesMatch(branch, package_name, vendor, product,
			package_relatives)
		if not ok:
			return False, msg

		# Considering versions that can be included in a package name and
		# a product name
		ok, msg = self.__VersionsInNamesMatch(package_name, product)
		if not ok:
			return False, msg

		return True, msg
