#!/usr/bin/python3

#	cve-manager : CVE management tool
#	Copyright (C) 2017-2022 Alexey Appolonov
#
#	This program is free software: you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation, either version 3 of the License, or
#	(at your option) any later version.
#
#	This program is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with this program.  If not, see <http://www.gnu.org/licenses/>.

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

import re
from copy                 import deepcopy
from mysql.connector      import Error as MySQLError
from collections          import defaultdict
from cve_manager.mediator import ProtoMediator
from cve_manager.common   import MATCH_VENDOR_PRODUCT
from cve_manager.conf     import LOCAL_SYS
from cve_manager.db       import DB, Placeholders
from cve_manager.url      import ParseAndJoinURL
from cpe_map              import defines as df

EMPTY_RESULT = 'Empty result'

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# An interface for interacting with the DB during making a mapping choice

class Mediator(ProtoMediator):

	def __init__(self, mysql_config, printer):

		super().__init__(mysql_config, printer)

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Get a condition and a corresponding data for selecting packages

	def __PackagesConditionAndData(self, selection=[]):

		if selection:
			condition = f'name in ({Placeholders(len(selection))})'
			data = [tuple(selection)]
		else:
			condition = ''
			data = []

		return condition, data

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Invert a given dict of sets

	def __Invert(self, d):

		inverted_d = defaultdict(set)

		for key, values in d.items():
			for v in values:
				inverted_d[v].add(key)

		return inverted_d

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Form a branch map for a given branch from a content of
	# a src_to_<data_source>_map merged with a content of
	# a bin_to_<data_source>_map

	def QueryBranchMap(self, branch, data_source, selection=[]):

		self.printer.LineBegin('Getting a map of package names '
			f'to {data_source.upper()} products')

		db = super().NewDB(2)
		if not db:
			return None

		src_table = f'{branch}_src'
		bin_table = f'{branch}_bin'
		src_map_table = f'src_to_{data_source}_map'
		bin_map_table = f'bin_to_{data_source}_map'

		packages, data = self.__PackagesConditionAndData(selection)

		# Forming a query to get a src map
		src_map_fields = [
			f'a.{branch}_{m}' if m in df.M_URL_EXPANDED or m == df.M_FIXES
			else f'a.{m}' for m in df.TM_SRC
			]
		src_map_query  = (
			f'SELECT a.package, b.url, {", ".join(src_map_fields)} '
			f'FROM {src_map_table} a INNER JOIN {src_table} b '
			f'ON a.package = b.name '
			+ (f'WHERE b.{packages}' if packages else '')
			)

		# Forming a query to get a bin map with bin package names replaced
		# with their src package names
		bin_map_fields = [f'a.{m}' for m in df.TM_BIN]
		bin_map_query  = (
			f'SELECT c.name, c.url, {", ".join(bin_map_fields)} '
			f'FROM {bin_map_table} a '
			f'INNER JOIN {bin_table} b ON a.package = b.name '
			f'INNER JOIN {src_table} c ON b.src_id = c.id '
			+ (f'WHERE c.{packages}' if packages else '')
			)

		# Merging the maps
		branch_map = defaultdict(lambda: defaultdict(dict))
		for query, table, i in [
				(src_map_query, src_map_table, 0),
				(bin_map_query, bin_map_table, 1),
				]:
			cursor = db.Cursor(query, tables=(table,), data=data, n=i)
			if not cursor:
				if cursor != None:
					self.printer.Err()
				return None
			for entry in cursor:
				package = entry[0]
				url = ParseAndJoinURL(entry[1])
				for j, tm in enumerate(df.TM_SRC if i == 0 else df.TM_BIN):
					items = entry[j + 2]
					if not items:
						continue
					# Matches of the type "invariant" contain no scores
					if tm in df.TM_INVARIANT:
						regexp = r'^' + MATCH_VENDOR_PRODUCT + '$'
						abs_score = True
					# Mathces of the type "variative" contain scores
					else:
						regexp = (r'^' + MATCH_VENDOR_PRODUCT + '\s'
							r'(?P<score>\d+\.?\d*)\s*$')
						abs_score = False
					for item in items.split('  '):
						m = re.match(regexp, item)
						if not m:
							continue
						vendor = m.group('vendor')[:-1] if m.group('vendor') else ''
						product = m.group('product')
						score = 1.0 if abs_score else float(m.group('score'))
						if branch_map[(package, url)][tm].get((vendor, product), 0.0) < score:
							branch_map[(package, url)][tm][(vendor, product)] = score

		if not branch_map:
			self.printer.Warn('Branch map is empty')
		else:
			self.printer.Success()

		return branch_map

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Getting a pair of dicts {vendor: product} of all not corrupted CPEs
	# (listed and free and not supplementary or not listed, but with a product
	# name that none of the corrupted CPEs has) and of all corrupted CPEs
	# (supplementary or not free); Supplementary CPE is one that is not in the
	# official CPE dict; Not listed CPE is one that 1) is in the official CPE
	# dict and 2) is not in the vulnerability lists and 3) whose product name
	# is same as product name of another CPE that is in the vulnerability lists

	def QueryNotCorruptedAndCorruptedProducts(self, type_of_products):

		CORRUPTED     = 0
		NOT_CORRUPTED = 1
		NOT_LISTED    = 2

		self.printer.LineBegin(f'Getting dicts of {type_of_products.upper()} '
			'products')

		db = super().NewDB()
		if not db:
			return None, None

		table = type_of_products + '_products'
		conditions = {
			CORRUPTED     : '(is_supplimentary = 1 OR  is_non_free = 1)',
			NOT_CORRUPTED : '(is_supplimentary = 0 AND is_non_free = 0 AND '
				'is_not_listed = 0)',
			NOT_LISTED    : 'is_not_listed = 1',
			}

		resp = defaultdict(lambda: defaultdict(set))
		for target, condition in conditions.items():
			cursor = db.Select(['vendor', 'product'], table, condition)
			if not cursor:
				self.printer.Warn(f'Can\'t get data from the "{table}" table')
				return None, None
			for vendor, product in cursor:
				resp[target][product].add(vendor)

		corrupted = resp[CORRUPTED]
		not_corrupted = resp[NOT_CORRUPTED]

		# Not listed CPE is not corrupted if there is no corrupted CPE with
		# a same product name
		for product, vendors in resp[NOT_LISTED].items():
			if product not in corrupted:
				not_corrupted[product] |= vendors

		if not corrupted and not not_corrupted:
			self.printer.Warn('Formed dicts are empty')
		else:
			self.printer.Success()

		return self.__Invert(not_corrupted), self.__Invert(corrupted)

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Get URLs of the specified packages from a specified table

	def __QueryURLs(self, table, packages):

		self.printer.LineBegin('Getting URLs of the presented packages')

		db = super().NewDB()
		if not db:
			return None

		cond, data = self.__PackagesConditionAndData(packages)
		cond = "url<>'' " + (f'AND {cond}' if cond else '')

		if not db.Select(['name', 'url'], table, cond, data):
			self.printer.Warn(EMPTY_RESULT)
			return {}

		urls = {package_name: ParseAndJoinURL(package_url)
			for package_name, package_url in db.cursor}

		if not urls:
			self.printer.Warn(EMPTY_RESULT)
			return {}

		self.printer.Success()

		return urls

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Get names of relatives of each package/product that has them

	def __QueryRelatives(self, msg, query, table, data=[]):

		self.printer.LineBegin(msg)

		db = super().NewDB()
		if not db:
			return None

		if not db.Cursor(query, tables=(table,), data=data):
			self.printer.Warn(EMPTY_RESULT)
			return {}

		relatives = defaultdict(list)

		# Forming a dict of ralated product/package names
		for name, related_names in db.cursor:
			# Relative with shortest name will be first in the list
			# - it's best to look upon this relative
			list_of_related_names = related_names.split()
			list_of_related_names.sort(key = lambda x: len(x))
			relatives[name] = list_of_related_names

		self.printer.Success()

		return relatives

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Get names of relatives of each package that has them

	def QueryRelatedPackages(self, branch, selection=[]):

		msg = 'Getting related packages of ' + \
			('the local system' if branch == LOCAL_SYS else f'{branch} branch')
		table = f'{branch}_src'
		condition, data = self.__PackagesConditionAndData(selection)
		query = f"SELECT name, relatives FROM {table} WHERE relatives<>'' " + \
			(f'AND {condition}' if condition else '')

		relatives = self.__QueryRelatives(msg, query, table, data)

		presented_packages = relatives.keys() | \
			{package for packages in relatives.values() for package in packages}

		urls = self.__QueryURLs(table, presented_packages)

		return {(package_name, urls.get(package_name, '')):
			[(rpackage_name, urls.get(rpackage_name, ''))
				for rpackage_name in rpackage_names]
			for package_name, rpackage_names in relatives.items()}

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Get names of relatives of each product that has them

	def QueryRelatedProducts(self, type_of_products):

		msg   = f'Getting relatives of {type_of_products.upper()} products'
		table = f'{type_of_products}_products'
		query = f"SELECT CONCAT(vendor, ':', product), relatives FROM {table} " \
			"WHERE relatives<>''"

		return self.__QueryRelatives(msg, query, table)

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Get a dict of current package versions

	def QueryCurrentPackageVersions(self, branch):

		msg = 'Getting current versions of ' \
			f'{"local system" if branch == LOCAL_SYS else branch} packages'
		self.printer.LineBegin(msg)

		if not branch:
			self.printer.Err('Branch name is not specified')
			return None

		db = super().NewDB()
		if not db:
			return None

		cursor = db.Select(['name', 'ver'], f'{branch}_src')
		res = {name: ver for name, ver in cursor} if cursor else None

		if not res:
			self.printer.Warn(EMPTY_RESULT)
		else:
			self.printer.Success()

		return res

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Get a dict of time lines of packages or products

	def QueryTimeLines(self, branch='', data_source=''):

		self.printer.LineBegin('Getting time lines of ' + \
			(f'{branch} packages' if branch
				else f'{data_source.upper()} products'))

		if not branch and not data_source:
			self.printer.Err('Need either "data_source" param or "branch" param')
			return None

		db = super().NewDB()
		if not db:
			return None

		if branch:
			# Getting a {name, version, timestamp} of every package
			# that were ever committed to this branch
			table  = f'{branch}_timelines'
			query  = f'SELECT name, ver, timestamp FROM {table}'
			tables = (table,)
		else:
			# Getting a {vendor, product, version, first mention date} of every
			# considered product
			table_a = f'{data_source}_products'
			table_b = f'{table_a}_timelines'
			query = 'SELECT a.vendor, a.product, b.ver, b.first_mention ' \
				f'FROM {table_a} a INNER JOIN {table_b} b USING (id)'
			tables = (table_a, table_b)

		cursor = db.Cursor(query, tables=tables)
		if not cursor:
			self.printer.Warn(EMPTY_RESULT)
			return {}

		if branch:
			time_lines = defaultdict(list)
			for name, ver, date in cursor:
				time_lines[name].append((ver, date))
		else:
			time_lines = defaultdict(lambda: defaultdict(list))
			for vendor, product, ver, date in cursor:
				time_lines[vendor][product].append((ver, date))

		self.printer.Success()

		return time_lines

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Send CPE/FSTEC map to the database

	def SendMap(self, choices, branch, target):

		self.printer.LineBegin(f'Sending {branch} to {target.upper()} '
			'mapping choices')

		if not choices:
			self.printer.Err('There is nothing to send')
			return False

		db = super().NewDB()
		if not db:
			return False

		table = f'{branch}_src'
		query = f'UPDATE {table} SET {target}_name = %s WHERE name = %s'
		data  = []
		for package, product in choices.items():
			data.append((product, package))

		if db.Cursor(query, data=data) == None:
			return False

		self.printer.Success()

		return True
