#!/usr/bin/python3

#	cve-manager : CVE management tool
#	Copyright (C) 2017-2022 Alexey Appolonov
#
#	This program is free software: you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation, either version 3 of the License, or
#	(at your option) any later version.
#
#	This program is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with this program.  If not, see <http://www.gnu.org/licenses/>.

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

from collections           import defaultdict
from copy                  import deepcopy
from math                  import exp, log10
from cpe_map               import defines as df
from cpe_map_choice.filter import Filter

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Sort values of a given dict and join their elements together into strings

def JoinValues(d):

	return {k: ' '.join(sorted(v)) for k, v in d.items()}

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

class Solver():

	def __init__(self, printer, verbose):

		self.RANKING = {
			df.M_PRESCRIBED:   100.0,
			df.M_URL_GRADE_A:  15.5,
			df.M_COMPLETE:     15.0,
			df.M_PARTIAL:      10.0,
			df.M_FIXES:        10.0,
			df.M_URL_GRADE_B:  9.5,
			df.M_BIN_COMPLETE: 1.0,
			df.M_BIN_PARTIAL:  1.0,
			}
		self.MIN_SCORE = 0.8 * self.RANKING[df.M_PARTIAL]
		self.ACCURACY  = 0.001
		self.SUPPL_PENALTY = 0.9 # Penalty for not being in official CPE dict
		self.printer = printer
		self.verbose = verbose
		self.filter  = None

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Extract prescriptions from a given branch map

	def __GetPrescriptions(self, branch_map):

		prescriptions = defaultdict(set)

		for package_name_and_url, matches in branch_map.items():
			for vendor_and_product in matches.get('prescribed', {}):
				prescriptions[vendor_and_product].add(package_name_and_url[0])

		return prescriptions

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Complement a dict of relatives with packages that are prescribed to the
	# same products

	def __Complement(self, relatives, prescriptions):

		res = deepcopy(relatives)

		for prescribed_packages in prescriptions.values():
			if len(prescribed_packages) < 2:
				continue
			fmt_prescribed_packages = {(p, '') for p in prescribed_packages}
			for prescribed_package in prescribed_packages:
				k = (prescribed_package, '')
				current_relatives = set(res.get(k, []))
				res[k] = \
					list((current_relatives | fmt_prescribed_packages) - {k})
				res[k].sort(key = lambda x: len(x[0]))

		return res

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Filter given branch map

	def __FilterBranchMap(self, branch, branch_map, related_package_names,
			allow_empty_result=False):

		self.printer.LineBegin(f'Filtering a branch map')

		if not self.filter:
			self.printer.Err('Filter obj has not been initialized')
			return None

		filtered_branch_map = defaultdict(lambda: defaultdict(dict))
		dropped_pairs = set()
		for package_name_and_url, matches in branch_map.items():
			package_name, package_url = package_name_and_url
			for tm, matches_for_tm in matches.items():
				package_rels = related_package_names.get(package_name_and_url, [])
				for vendor_and_product, score in matches_for_tm.items():
					ok, msg = self.filter.Pass(branch, package_name, package_url,
						*vendor_and_product, package_rels, tm)
					if ok:
						filtered_branch_map[package_name][tm][vendor_and_product] = score
					elif (package_name_and_url, vendor_and_product) in dropped_pairs:
						continue
					# Eliminating repetitive messages
					dropped_pairs.add((package_name_and_url, vendor_and_product))
					if msg and self.verbose:
						self.printer.LineAddExtra(msg)

		if not filtered_branch_map:
			msg = 'Empty result'
			if allow_empty_result:
				self.printer.Warn(msg)
			else:
				self.printer.Err(msg)
		else:
			self.printer.Success()

		return filtered_branch_map

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Get rank of the particular type of matching

	def __GetRank(self, type_of_matching):

		rank = 0

		for tm in type_of_matching.split('+'):
			rank += self.RANKING.get(tm, 0)

		return rank

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Normalize criterion values of a branch map

	def __NormBranchMap(self, branch_map):

		norm_branch_map = deepcopy(branch_map)

		# Calculate norm k for a given type of matching of a given branch map
		def NormK(branch_map, type_of_matching):
			max_crit = 0
			# Packages are the keys
			for matches in branch_map.values():
				matches_for_tm = matches.get(type_of_matching, {})
				for score in matches_for_tm.values():
					if max_crit < score:
						max_crit = score
			return max_crit
		# - nested func

		# Get magnitude of a ranking of a given type of matching
		def RkMagnitude(type_of_matching):
			return int(log10(self.RANKING.get(type_of_matching, 0)))
		# - nested func

		# Performing a linear normalization (for corresp. types of matching)
		for type_of_matching in df.TM_VARIATIVE_NOT_NORM_LIN:
			norm_k = NormK(branch_map, type_of_matching)
			rank_m = Magnitude(type_of_matching)
			if norm_k <= 0:
				continue
			# Packages are the keys
			for matches in norm_branch_map.values():
				matches_for_tm = matches.get(type_of_matching, {})
				for product, score in matches_for_tm.items():
					matches_for_tm[product] = round(score/norm_k, rank_m)

		# Performing a normalization with the use of exponential func (--//--)
		for type_of_matching in df.TM_VARIATIVE_NOT_NORM_EXP:
			rank_m = RkMagnitude(type_of_matching)
			# Packages are the keys
			for matches in norm_branch_map.values():
				matches_for_tm = matches.get(type_of_matching, {})
				for product, score in matches_for_tm.items():
					matches_for_tm[product] = round(1 - 1 / exp(score / 3), rank_m)

		return norm_branch_map

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Get a dict of sum scores for given combinations of types of matching

	def __IntegralScores(self, norm_branch_map, combinations,
			supplementary_products, supplementary_products_flat):

		criterions = defaultdict(lambda: defaultdict(lambda: defaultdict(float)))

		# Forming a dict of all variants of products for each package name
		products = defaultdict(set)
		for package, matches in norm_branch_map.items():
			for _, matches_for_tm in matches.items():
				for vendor_and_product in matches_for_tm:
					products[package].add(vendor_and_product)

		# Define a value of a penalty coefficient
		def GetPenalty(type_of_matching, vendor, product):
			if type_of_matching in \
					(df.M_URL_GRADE_A, df.M_URL_GRADE_B, df.M_FIXES):
				if product in supplementary_products.get(vendor, {}):
					return self.SUPPL_PENALTY
			elif type_of_matching != df.M_PRESCRIBED:
				if product in supplementary_products_flat:
					return self.SUPPL_PENALTY
			return 1
		# - nested func

		for combination in combinations:
			if not combination or \
					any(tm not in df.M_CHOICES_EXPANDED for tm in combination):
				continue
			integral_tm = '+'.join(combination)
			# Summing the scores of various types of matching
			for package, matches in norm_branch_map.items():
				if any(tm not in df.M_CHOICES_EXPANDED for tm in matches.keys()):
					continue
				for tm in combination:
					matches_for_tm = matches.get(tm, {})
					scores_for_tm = {}
					for vendor_and_product, score in matches_for_tm.items():
						if score < self.ACCURACY:
							continue
						# Considering a rank of a type of matching and a penalty
						#	coefficient
						scores_for_tm[vendor_and_product] = score * \
							self.__GetRank(tm) * \
							GetPenalty(tm, *vendor_and_product)
					# Adding scores for this type of matching while considering
					#	scores for matches with no vendor name, just product name
					for vendor_and_product, score in scores_for_tm.items():
						criterions[package][integral_tm][vendor_and_product] += score
						vendor, product = vendor_and_product
						if vendor:
							continue
						for vendor_and_product_integral in products.get(package, []):
							vendor_integral, product_integral = vendor_and_product_integral
							if not vendor_integral or \
									product_integral != product or \
									scores_for_tm.get(vendor_and_product_integral):
								continue
							criterions[package][integral_tm][vendor_and_product_integral] += score

		return criterions

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Check if a score of a new product is higher than a score of a currently
	# chosen product

	def __NewScoreIsHigher(self, new_product, new_score,
			cur_product, cur_score):

		if not cur_product or not cur_score:
			return True

		# If score of a current choice for this package is no less
		if cur_score >= (new_score + self.ACCURACY):
			return False

		# If scores are almost equal
		if new_score <= (cur_score + self.ACCURACY):

			if cur_product == new_product:
				return False

			# Mapping of multiple product names to a single package name
			# is allowed
			if new_score >= (self.RANKING[df.M_PRESCRIBED] - self.ACCURACY):
				return True

			# Picking the one that is earlier in the alphabet
			if new_product > cur_product:
				return False

		return True

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Make choices by determining a greatest score for each package and
	# a greatest score for each product

	def __Choices(self, best_scores, related_packages):

		choices = defaultdict(set)
		reverse_choices = defaultdict(set)
		sorted_packages = sorted(best_scores)
		related_package_names = {k[0]: [pair[0] for pair in v] \
			for k, v in related_packages.items()}

		while True:

			new_choice_made = False

			for package in sorted_packages:

				matches = best_scores[package]

				for product, score in matches.items():

					# Checking if a score of this product is higher than a score
					# of a currently chosen product
					if any(not self.__NewScoreIsHigher(product, score, p, s)
							for p, s in choices.get(package, {(None, None)})):
						continue
					# Getting a current set of matches for this product
					choices_for_this_product = reverse_choices.get(product, set())
					# Checking if a package is not a relative to all packages
					# that are matched to this product
					packages_are_not_related = \
						any(package not in related_package_names.get(name, [])
							for name, _ in choices_for_this_product)
					# If this product has a higher score for a match
					# with other package
					if any(score <= s for _, s in choices_for_this_product) and \
							packages_are_not_related:
						continue
					# Current match will become a choice for this package
					new_choice_made = True
					# Canceling choices for packages that were previously mapped
					# to this product
					for name, _ in choices_for_this_product:
						if name not in related_package_names.get(package, []):
							choices.pop(name)
					# Canceling choices for this product
					if packages_are_not_related:
						reverse_choices[product] = set()
					# Making a new choice (mapping of multiple product names
					# to a single package name is allowed only for prescribed
					# matches)
					if score >= (self.RANKING[df.M_PRESCRIBED] - self.ACCURACY):
						choices[package].add((product, score))
					else:
						choices[package] = {(product, score)}
					reverse_choices[product].add((package, score))

				#  Filtering irrelevant reverse choices
				chosen_products = {p for p, _ in choices.get(package, {})}
				if not chosen_products:
					continue
				buf = {}
				for product, choices_for_this_product in reverse_choices.items():
					for name, _ in choices_for_this_product:
						if name == package and product not in chosen_products:
							buf[product] = []
							break
				for product in buf:
					for name, s in reverse_choices[product]:
						if name != package:
							buf[product].append((name, s))
				for product, entries in buf.items():
					if entries:
						reverse_choices[product] = entries
					else:
						del reverse_choices[product]

			# If no new pairs were found on a last iteration over a whole dict
			# of scores
			if not new_choice_made:
				break

		# Throwing away the scores
		res = defaultdict(set)
		res_reverse = defaultdict(set)
		for package, product_and_score_pairs in choices.items():
			for product, _ in product_and_score_pairs:
				res[package].add(product)
				res_reverse[product].add(package)

		return JoinValues(res), JoinValues(res_reverse)

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Make a choice based on choices made for related products

	def __ChoicesForRelatedProducts(self, branch, related_products,
			choices_reverse):

		choices_for_related_products = defaultdict(set)

		# Retrieve product name from a "<vendor>:<product>" string
		def GetProductName(vendor_and_product):
			try:
				_, product = vendor_and_product.split(':')
				return product
			except ValueError:
				self.printer.Err('Wrong format of string of related products '
					f'-- {vendor_and_product}')
			return None
		# - nested func

		for vendor_and_product, relatives_of_this_product in related_products.items():
			product = GetProductName(vendor_and_product)
			if product == None:
				return None
			if product in choices_reverse or not relatives_of_this_product:
				continue
			for related_vendor_and_product in relatives_of_this_product:
				related_product = GetProductName(related_vendor_and_product)
				if related_product == None:
					return None
				package_names = choices_reverse.get(related_product, '')
				for package_name in package_names.split():
					ok, msg = self.filter.Pass(branch, package_name, '', '', product)
					if not ok:
						continue
					choices_for_related_products[package_name].add(related_product)
					choices_for_related_products[package_name].add(product)

		return JoinValues(choices_for_related_products)

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Pick a best score for every match while rejecting scores that don't pass
	# the minimal limit

	def __BestScores(self, branch_map):

		best_scores = defaultdict(lambda: defaultdict(float))

		for package, matches in branch_map.items():
			for scores in matches.values():
				for vendor_and_product, score in scores.items():
					_, product = vendor_and_product
					if score >= (self.MIN_SCORE - self.ACCURACY) and \
							best_scores[package][product] < score:
						best_scores[package][product] = score

		return best_scores

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Make a mapping choice based on:
	# - Current versions of packages
	# - Current mapping table;
	# - Dict of related package names;
	#   {package: <list of package names that relate to this package name>};
	# - Dict of mapping pairs that should be ignored/skipped;
	# - Timelines of packages (time when each new ver of each pack was build);
	# - Timelines of products (time when each ver of a product was first
	#   mentioned in NVD data);
	# - Dict of related package names.

	def MakeChoice(self, branch, branch_map, related_packages, ignore={},
			supplementary_products={}, supplementary_products_flat=set(),
			current_versions_of_packages={},
			time_lines_of_packages={}, time_lines_of_products={},
			related_products={}, allow_empty_result=False):

		# Extracting prescriptions from a branch map
		prescriptions = self.__GetPrescriptions(branch_map)

		# Complementing a dict of related packages with package names that are
		# prescribed to the same products
		related_packages_complemented = \
			self.__Complement(related_packages, prescriptions)

		# Filtering and normalizing a branch map
		self.filter = Filter(ignore, current_versions_of_packages,
			time_lines_of_packages, time_lines_of_products)
		filtered_branch_map = self.__FilterBranchMap(
			branch, branch_map, related_packages_complemented, allow_empty_result)
		if filtered_branch_map == None:
			return None
		norm_branch_map = self.__NormBranchMap(filtered_branch_map)

		self.printer.LineBegin("Making a choice")

		# Calculating integral scores for a normalized branch map
		tm_incompatible_x = (
			df.M_COMPLETE,
			df.M_PARTIAL,
			df.M_BIN_COMPLETE,
			df.M_BIN_PARTIAL,
			)
		tm_incompatible_y = (
			df.M_URL_GRADE_A,
			df.M_URL_GRADE_B,
			)
		tm_combinations = ((df.M_PRESCRIBED,),) + \
			(tuple((tm_x, tm_y, df.M_FIXES,)
				for tm_x in tm_incompatible_x for tm_y in tm_incompatible_y))
		integral_scores = self.__IntegralScores(norm_branch_map, tm_combinations,
			supplementary_products, supplementary_products_flat)

		# Picking a best score for every match
		best_scores = self.__BestScores(integral_scores)

		# Making choices by greater scores
		choices, choices_reverse = \
			self.__Choices(best_scores, related_packages_complemented)

		# Making a choice based on choices made for related packages
		for package_name_and_url, relatives_of_this_package in \
				related_packages.items():
			package_name, package_url = package_name_and_url
			if package_name in choices or not relatives_of_this_package:
				continue
			for related_package_name, _ in relatives_of_this_package:
				products = choices.get(related_package_name, '')
				for product in products.split():
					ok, msg = self.filter.Pass(
						branch, package_name, package_url, '', product)
					if ok:
						choices[package_name] = product
					if msg and self.verbose:
						self.printer.LineAddExtra(msg)

		# Making a choice based on choices made for related products
		choices_for_related_products = self.__ChoicesForRelatedProducts(
			branch, related_products, choices_reverse)

		for package, products in choices_for_related_products.items():
			choices[package] = products

		if choices:
			n = len(choices)
			self.printer.Success(f'{n} mapped package{"s" if n > 1 else ""}')
		else:
			self.printer.Warn('None of the packages was mapped')

		return choices
